@@ -264,6 +264,8 @@ struct bch_sb *bch2_format(struct bch_opt_strs	fs_opt_strs,
 	SET_BCH_SB_META_REPLICAS_REQ(sb.sb, 1);
 	SET_BCH_SB_DATA_REPLICAS_REQ(sb.sb, 1);
 
+	SET_BCH_SB_EXTENT_BP_SHIFT(sb.sb, 16);
+
 	if (opts.version > bcachefs_metadata_version_disk_accounting_big_endian)
 		sb.sb->features[0] |= cpu_to_le64(BCH_SB_FEATURES_ALL);
 
@@ -296,7 +298,12 @@ struct bch_sb *bch2_format(struct bch_opt_strs	fs_opt_strs,
 		uuid_generate(m->uuid.b);
 		m->nbuckets	= cpu_to_le64(i->nbuckets);
 		m->first_bucket	= 0;
+
+		if (!opt_defined(i->opts, rotational))
+			opt_set(i->opts, rotational, !bdev_nonrot(i->bdev));
+
 		bch2_opt_set_sb_all(sb.sb, idx, &i->opts);
+		SET_BCH_MEMBER_ROTATIONAL_SET(m, true);
 	}
 
 	/* Disk labels*/
