@@ -73,6 +73,10 @@ void bch2_backpointer_to_text(struct printbuf *out, struct bch_fs *c, struct bke
 		   bp.v->bucket_len,
 		   bp.v->bucket_gen);
 	bch2_bpos_to_text(out, bp.v->pos);
+
+	if (BACKPOINTER_RECONCILE_PHYS(bp.v))
+		prt_str(out, " phys");
+
 }
 
 void bch2_backpointer_swab(const struct bch_fs *c, struct bkey_s k)
@@ -209,17 +213,17 @@ static int backpointer_target_not_found(struct btree_trans *trans,
 
 	prt_printf(&buf, "backpointer doesn't match %s it points to:\n",
 		   bp.v->level ? "btree node" : "extent");
-	bch2_bkey_val_to_text(&buf, c, bp.s_c);
-
-	prt_newline(&buf);
 	bch2_bkey_val_to_text(&buf, c, target_k);
 
+	prt_str(&buf, "\nfound: ");
+	bch2_bkey_val_to_text(&buf, c, bp.s_c);
+
 	struct bkey_ptrs_c ptrs = bch2_bkey_ptrs_c(target_k);
 	const union bch_extent_entry *entry;
 	struct extent_ptr_decoded p;
 	bkey_for_each_ptr_decode(target_k.k, ptrs, p, entry)
 		if (p.ptr.dev == bp.k->p.inode) {
-			prt_newline(&buf);
+			prt_str(&buf, "\nwant:  ");
 			struct bkey_i_backpointer bp2;
 			bch2_extent_ptr_to_bp(c, bp.v->btree_id, bp.v->level, target_k, p, entry, &bp2);
 			bch2_bkey_val_to_text(&buf, c, bkey_i_to_s_c(&bp2.k_i));
@@ -949,8 +953,13 @@ static int check_bucket_backpointer_mismatch(struct btree_trans *trans, struct b
 		 * Pre upgrade, we expect all the buckets to be wrong, a write
 		 * buffer flush is pointless:
 		 */
-		if (c->sb.version_upgrade_complete >= bcachefs_metadata_version_backpointer_bucket_gen)
+		if (c->sb.version_upgrade_complete >= bcachefs_metadata_version_backpointer_bucket_gen) {
+			if (a->data_type == BCH_DATA_btree) {
+				bch2_trans_unlock_long(trans);
+				bch2_btree_interior_updates_flush(c);
+			}
 			try(bch2_backpointers_maybe_flush(trans, alloc_k, last_flushed));
+		}
 
 		bool empty = (sectors[ALLOC_dirty] +
 			      sectors[ALLOC_stripe] +
@@ -1124,8 +1133,10 @@ int bch2_check_extents_to_backpointers(struct bch_fs *c)
 		nr_empty	+= ca->bucket_backpointer_empty.nr;
 	}
 
+#ifndef CONFIG_BCACHEFS_DEBUG
 	if (!nr_mismatches)
 		goto err;
+#endif
 
 	bch_info(c, "scanning for missing backpointers in %llu/%llu buckets, %llu buckets with no backpointers",
 		 nr_mismatches - nr_empty, nr_buckets, nr_empty);
@@ -1357,6 +1368,9 @@ struct bkey_s_c_backpointer bch2_bp_scan_iter_peek(struct btree_trans *trans,
 			if (k.k->type != KEY_TYPE_backpointer)
 				continue;
 
+			/* XXX: this is a really big allocation, we should drop
+			 * srcu lock */
+
 			struct bkey_i_backpointer bp;
 			bkey_reassemble(&bp.k_i, k);
 			if (iter->bps.nr > limit ||
@@ -1368,7 +1382,6 @@ struct bkey_s_c_backpointer bch2_bp_scan_iter_peek(struct btree_trans *trans,
 			 ? bch2_progress_update_iter(trans, iter->progress, &bp_iter)
 			 : 0);
 		}));
-		trans->restart_count = restart_count;
 
 		if (ret)
 			return ((struct bkey_s_c_backpointer) { .k = ERR_PTR(ret) });
@@ -1376,7 +1389,10 @@ struct bkey_s_c_backpointer bch2_bp_scan_iter_peek(struct btree_trans *trans,
 		if (!iter->bps.nr)
 			return (struct bkey_s_c_backpointer) {};
 
+		bch2_trans_unlock_long(trans);
 		darray_sort(iter->bps, bkey_i_backpointer_cmp);
+		bch2_trans_begin(trans);
+		trans->restart_count = restart_count;
 	}
 
 	return backpointer_i_to_s_c(&darray_last(iter->bps));
