@@ -907,7 +907,7 @@ static bool should_drop_bucket(struct open_bucket *ob, struct bch_fs *c,
 
 		if (!drop && ob->ec) {
 			guard(mutex)(&ob->ec->lock);
-			unsigned nr_blocks = bkey_i_to_stripe(&ob->ec->new_stripe.key)->v.nr_blocks;
+			unsigned nr_blocks = ob->ec->new_stripe.key.v.nr_blocks;
 
 			for (unsigned i = 0; i < nr_blocks; i++) {
 				if (!ob->ec->blocks[i])
@@ -1163,7 +1163,7 @@ retry:
 	req->will_retry_target_devices	= !(req->flags & BCH_WRITE_alloc_nowait);
 	req->ptrs.nr			= 0;
 	req->nr_effective		= 0;
-	req->have_cache			= false;
+	req->have_cache			= req->flags & BCH_WRITE_move;
 	write_points_nr			= a->write_points_nr;
 
 	*wp_ret = req->wp = writepoint_find(trans, write_point.v);
