@@ -9,6 +9,7 @@
 #include "bcachefs.h"
 
 #include "alloc/buckets.h"
+#include "alloc/disk_groups.h"
 
 #include "btree/bkey_methods.h"
 #include "btree/cache.h"
@@ -19,7 +20,7 @@
 #include "data/checksum.h"
 #include "data/compress.h"
 #include "data/extents.h"
-#include "data/rebalance.h"
+#include "data/reconcile/trigger.h"
 
 #include "fs/inode.h"
 
@@ -28,6 +29,8 @@
 #include "util/util.h"
 
 #ifdef CONFIG_BCACHEFS_DEBUG
+#include <linux/module.h>
+
 static int bch2_force_read_device = -1;
 
 module_param_named(force_read_device, bch2_force_read_device, int, 0644);
@@ -55,19 +58,7 @@ void bch2_io_failures_to_text(struct printbuf *out,
 			      struct bch_fs *c,
 			      struct bch_io_failures *failed)
 {
-	static const char * const error_types[] = {
-		"btree validate", "io", "checksum", "ec reconstruct", NULL
-	};
-
-	for (struct bch_dev_io_failures *f = failed->devs;
-	     f < failed->devs + failed->nr;
-	     f++) {
-		unsigned errflags =
-			((!!f->failed_btree_validate)	<< 0) |
-			((!!f->failed_io)		<< 1) |
-			((!!f->failed_csum_nr)		<< 2) |
-			((!!f->failed_ec)		<< 3);
-
+	darray_for_each(*failed, f) {
 		bch2_printbuf_make_room(out, 1024);
 		scoped_guard(rcu) {
 			guard(printbuf_atomic)(out);
@@ -78,69 +69,54 @@ void bch2_io_failures_to_text(struct printbuf *out,
 				prt_printf(out, "(invalid device %u)", f->dev);
 		}
 
-		prt_char(out, ' ');
+		if (!f->csum_nr && !f->ec_errcode && !f->errcode)
+			prt_str(out, " no error - confused");
 
-		if (!errflags) {
-			prt_str(out, "no error - confused");
-		} else if (is_power_of_2(errflags)) {
-			prt_bitflags(out, error_types, errflags);
-			prt_str(out, " error");
-		} else {
-			prt_str(out, "errors: ");
-			prt_bitflags(out, error_types, errflags);
-		}
+		if (f->csum_nr)
+			prt_printf(out, " checksum (%u)", f->csum_nr);
+		if (f->errcode)
+			prt_printf(out, " %s", bch2_err_str(f->errcode));
 		prt_newline(out);
+		if (f->ec_errcode)
+			prt_printf(out, "  ec reconstruct %s\n", bch2_err_str(f->ec_errcode));
 	}
 }
 
 struct bch_dev_io_failures *bch2_dev_io_failures(struct bch_io_failures *f,
 						 unsigned dev)
 {
-	struct bch_dev_io_failures *i;
-
-	for (i = f->devs; i < f->devs + f->nr; i++)
-		if (i->dev == dev)
-			return i;
-
-	return NULL;
+	return darray_find_p(*f, i, i->dev == dev);
 }
 
-void bch2_mark_io_failure(struct bch_io_failures *failed,
-			  struct extent_ptr_decoded *p,
-			  bool csum_error)
-{
-	struct bch_dev_io_failures *f = bch2_dev_io_failures(failed, p->ptr.dev);
-
-	if (!f) {
-		BUG_ON(failed->nr >= ARRAY_SIZE(failed->devs));
-
-		f = &failed->devs[failed->nr++];
-		memset(f, 0, sizeof(*f));
-		f->dev = p->ptr.dev;
-	}
-
-	if (p->do_ec_reconstruct)
-		f->failed_ec = true;
-	else if (!csum_error)
-		f->failed_io = true;
-	else
-		f->failed_csum_nr++;
-}
-
-void bch2_mark_btree_validate_failure(struct bch_io_failures *failed,
-				      unsigned dev)
+struct bch_dev_io_failures *bch2_dev_io_failures_mut(struct bch_io_failures *failed,
+						     unsigned dev)
 {
 	struct bch_dev_io_failures *f = bch2_dev_io_failures(failed, dev);
-
 	if (!f) {
-		BUG_ON(failed->nr >= ARRAY_SIZE(failed->devs));
+		BUG_ON(failed->nr >= ARRAY_SIZE(failed->data));
 
-		f = &failed->devs[failed->nr++];
+		f = &failed->data[failed->nr++];
 		memset(f, 0, sizeof(*f));
 		f->dev = dev;
 	}
 
-	f->failed_btree_validate = true;
+	return f;
+}
+
+void bch2_mark_io_failure(struct bch_io_failures *failed,
+			  struct extent_ptr_decoded *p, int err)
+{
+	BUG_ON(!err);
+	BUG_ON(bch2_err_matches(err, BCH_ERR_transaction_restart));
+
+	struct bch_dev_io_failures *f = bch2_dev_io_failures_mut(failed, p->ptr.dev);
+
+	if (p->do_ec_reconstruct)
+		f->ec_errcode = err;
+	else if (err == -BCH_ERR_data_read_retry_csum_err)
+		f->csum_nr++;
+	else
+		f->errcode = err;
 }
 
 static inline u64 dev_latency(struct bch_dev *ca)
@@ -150,7 +126,7 @@ static inline u64 dev_latency(struct bch_dev *ca)
 
 static inline int dev_failed(struct bch_dev *ca)
 {
-	return !ca || ca->mi.state == BCH_MEMBER_STATE_failed;
+	return !ca || ca->mi.state == BCH_MEMBER_STATE_evacuating;
 }
 
 /*
@@ -159,16 +135,11 @@ static inline int dev_failed(struct bch_dev *ca)
 static inline bool ptr_better(struct bch_fs *c,
 			      const struct extent_ptr_decoded p1,
 			      u64 p1_latency,
-			      struct bch_dev *ca1,
 			      const struct extent_ptr_decoded p2,
-			      u64 p2_latency)
+			      u64 p2_latency,
+			      unsigned preferred_dev,
+			      enum bch_read_flags flags)
 {
-	struct bch_dev *ca2 = bch2_dev_rcu_noerror(c, p2.ptr.dev);
-
-	int failed_delta = dev_failed(ca1) - dev_failed(ca2);
-	if (unlikely(failed_delta))
-		return failed_delta < 0;
-
 	if (static_branch_unlikely(&bch2_force_reconstruct_read))
 		return p1.do_ec_reconstruct > p2.do_ec_reconstruct;
 
@@ -194,6 +165,19 @@ static inline bool ptr_better(struct bch_fs *c,
 	if (unlikely(delta))
 		return delta > 0;
 
+	if (unlikely(flags & BCH_READ_soft_require_read_device)) {
+		if (p1.ptr.dev == preferred_dev) {
+			if (bch2_dev_rotational(c, p2.ptr.dev))
+				return true;
+			p1_latency /= 256;
+		}
+		if (p2.ptr.dev == preferred_dev) {
+			if (bch2_dev_rotational(c, p1.ptr.dev))
+				return false;
+			p2_latency /= 256;
+		}
+	}
+
 	/* Pick at random, biased in favor of the faster device: */
 
 	return bch2_get_random_u64_below(p1_latency + p2_latency) > p1_latency;
@@ -207,7 +191,8 @@ static inline bool ptr_better(struct bch_fs *c,
 int bch2_bkey_pick_read_device(struct bch_fs *c, struct bkey_s_c k,
 			       struct bch_io_failures *failed,
 			       struct extent_ptr_decoded *pick,
-			       int dev)
+			       unsigned preferred_dev,
+			       enum bch_read_flags flags)
 {
 	bool have_csum_errors = false, have_io_errors = false, have_missing_devs = false;
 	bool have_dirty_ptrs = false, have_pick = false;
@@ -234,7 +219,8 @@ int bch2_bkey_pick_read_device(struct bch_fs *c, struct bkey_s_c k,
 		}
 
 		/* Are we being asked to read from a specific device? */
-		if (dev >= 0 && p.ptr.dev != dev)
+		if ((flags & BCH_READ_hard_require_read_device) &&
+		    p.ptr.dev != preferred_dev)
 			continue;
 
 		struct bch_dev *ca = bch2_dev_rcu_noerror(c, p.ptr.dev);
@@ -251,21 +237,19 @@ int bch2_bkey_pick_read_device(struct bch_fs *c, struct bkey_s_c k,
 		struct bch_dev_io_failures *f =
 			unlikely(failed) ? bch2_dev_io_failures(failed, p.ptr.dev) : NULL;
 		if (unlikely(f)) {
-			p.crc_retry_nr	   = f->failed_csum_nr;
-			p.has_ec	  &= ~f->failed_ec;
+			p.crc_retry_nr	   = f->csum_nr;
+			p.has_ec	  &= !f->ec_errcode;
 
-			if (ca && ca->mi.state != BCH_MEMBER_STATE_failed) {
-				have_io_errors	|= f->failed_io;
-				have_io_errors	|= f->failed_btree_validate;
-				have_io_errors	|= f->failed_ec;
+			if (ca) {
+				have_io_errors	|= f->errcode != 0;
+				have_io_errors	|= f->ec_errcode;
 			}
-			have_csum_errors	|= !!f->failed_csum_nr;
+			have_csum_errors	|= f->csum_nr != 0;
 
-			if (p.has_ec && (f->failed_io || f->failed_csum_nr))
+			if (p.has_ec && (f->errcode || f->csum_nr))
 				p.do_ec_reconstruct = true;
-			else if (f->failed_io ||
-				 f->failed_btree_validate ||
-				 f->failed_csum_nr > c->opts.checksum_err_retry_nr)
+			else if (f->errcode ||
+				 f->csum_nr > c->opts.checksum_err_retry_nr)
 				continue;
 		}
 
@@ -290,8 +274,9 @@ int bch2_bkey_pick_read_device(struct bch_fs *c, struct bkey_s_c k,
 
 		if (!have_pick ||
 		    ptr_better(c,
-			       p, p_latency, ca,
-			       *pick, pick_latency)) {
+			       p, p_latency,
+			       *pick, pick_latency,
+			       preferred_dev, flags)) {
 			*pick = p;
 			pick_latency = p_latency;
 			have_pick = true;
@@ -627,8 +612,7 @@ bool bch2_bkey_narrow_crc(const struct bch_fs *c,
 {
 	BUG_ON(crc_is_compressed(new));
 	BUG_ON(new.offset);
-	BUG_ON(new.live_size != k->k.size);
-
+	BUG_ON(new.uncompressed_size != new.live_size);
 
 	union bch_extent_entry *old_e = bkey_crc_find(c, k, old);
 	if (!old_e)
@@ -796,7 +780,7 @@ void bch2_bkey_propagate_incompressible(const struct bch_fs *c, struct bkey_i *d
 
 	/*
 	 * XXX: if some data actually is compressed, we want
-	 * bch_extent_rebalance.wont_recompress_smaller
+	 * bch_extent_reconcile.wont_recompress_smaller
 	 */
 
 	bkey_extent_entry_for_each(ptrs, entry) {
@@ -840,34 +824,110 @@ unsigned bch2_bkey_replicas(struct bch_fs *c, struct bkey_s_c k)
 	return replicas;
 }
 
-unsigned bch2_extent_ptr_desired_durability(struct bch_fs *c, struct extent_ptr_decoded *p)
+unsigned bch2_dev_durability(struct bch_fs *c, unsigned dev)
 {
-	struct bch_dev *ca = bch2_dev_rcu_noerror(c, p->ptr.dev);
+	guard(rcu)();
+	struct bch_dev *ca = bch2_dev_rcu_noerror(c, dev);
+
+	return ca && ca->mi.state != BCH_MEMBER_STATE_evacuating
+		? ca->mi.durability
+		: 0;
+}
+
+static inline unsigned __extent_ptr_durability(struct bch_dev *ca, struct extent_ptr_decoded *p)
+{
+	if (p->ptr.cached)
+		return 0;
+
+	return p->has_ec
+		? p->ec.redundancy + 1
+		: ca->mi.durability;
+}
+
+int bch2_extent_ptr_desired_durability(struct btree_trans *trans, struct extent_ptr_decoded *p)
+{
+	guard(rcu)();
+	struct bch_dev *ca = bch2_dev_rcu_noerror(trans->c, p->ptr.dev);
 
 	return ca ? __extent_ptr_durability(ca, p) : 0;
 }
 
-unsigned bch2_extent_ptr_durability(struct bch_fs *c, struct extent_ptr_decoded *p)
+static unsigned bch2_extent_ptr_durability_rcu(struct bch_fs *c, struct extent_ptr_decoded *p)
 {
 	struct bch_dev *ca = bch2_dev_rcu_noerror(c, p->ptr.dev);
 
-	if (!ca || ca->mi.state == BCH_MEMBER_STATE_failed)
+	if (!ca || ca->mi.state == BCH_MEMBER_STATE_evacuating)
 		return 0;
 
 	return __extent_ptr_durability(ca, p);
 }
 
-unsigned bch2_bkey_durability(struct bch_fs *c, struct bkey_s_c k)
+int bch2_extent_ptr_durability(struct btree_trans *trans, struct extent_ptr_decoded *p)
+{
+	guard(rcu)();
+	return bch2_extent_ptr_durability_rcu(trans->c, p);
+}
+
+int bch2_bkey_durability(struct btree_trans *trans, struct bkey_s_c k, struct bkey_durability *ret)
+{
+	struct bch_fs *c = trans->c;
+	struct bkey_ptrs_c ptrs = bch2_bkey_ptrs_c(k);
+	const union bch_extent_entry *entry;
+	struct extent_ptr_decoded p;
+
+	*ret = (struct bkey_durability) {};
+
+	bkey_for_each_ptr_decode(k.k, ptrs, p, entry) {
+		if (p.ptr.dev == BCH_SB_MEMBER_INVALID)
+			continue;
+
+		int d = bch2_extent_ptr_durability(trans, &p);
+		if (d < 0)
+			return d;
+
+		if (bch2_dev_idx_is_online(c, p.ptr.dev))
+			ret->online += d;
+		ret->total += d;
+	}
+	return 0;
+}
+
+struct bkey_durability bch2_btree_ptr_durability(struct bch_fs *c, struct bkey_s_c k)
+{
+	BUG_ON(!bkey_is_btree_ptr(k.k));
+
+	struct bkey_ptrs_c ptrs = bch2_bkey_ptrs_c(k);
+	const union bch_extent_entry *entry;
+	struct extent_ptr_decoded p;
+	struct bkey_durability ret = {};
+
+	guard(rcu)();
+	bkey_for_each_ptr_decode(k.k, ptrs, p, entry) {
+		if (p.ptr.dev == BCH_SB_MEMBER_INVALID)
+			continue;
+
+		unsigned d = bch2_extent_ptr_durability_rcu(c, &p);
+
+		if (test_bit(p.ptr.dev, c->devs_online.d))
+			ret.online += d;
+		ret.total += d;
+	}
+	return ret;
+}
+
+bool bch2_bkey_can_read(const struct bch_fs *c, struct bkey_s_c k)
 {
 	struct bkey_ptrs_c ptrs = bch2_bkey_ptrs_c(k);
 	const union bch_extent_entry *entry;
 	struct extent_ptr_decoded p;
-	unsigned durability = 0;
 
-	guard(rcu)();
 	bkey_for_each_ptr_decode(k.k, ptrs, p, entry)
-		durability += bch2_extent_ptr_durability(c, &p);
-	return durability;
+		if (!p.ptr.cached &&
+		    (p.ptr.dev != BCH_SB_MEMBER_INVALID ||
+		     p.has_ec))
+			return true;
+
+	return false;
 }
 
 static unsigned bch2_bkey_durability_safe(struct bch_fs *c, struct bkey_s_c k)
@@ -880,10 +940,19 @@ static unsigned bch2_bkey_durability_safe(struct bch_fs *c, struct bkey_s_c k)
 	guard(rcu)();
 	bkey_for_each_ptr_decode(k.k, ptrs, p, entry)
 		if (p.ptr.dev < c->sb.nr_devices && c->devs[p.ptr.dev])
-			durability += bch2_extent_ptr_durability(c, &p);
+			durability += bch2_extent_ptr_durability_rcu(c, &p);
 	return durability;
 }
 
+void bch2_bkey_extent_entry_drop_s(const struct bch_fs *c, struct bkey_s k, union bch_extent_entry *entry)
+{
+	union bch_extent_entry *end = bkey_val_end(k);
+	union bch_extent_entry *next = extent_entry_next(c, entry);
+
+	memmove_u64s(entry, next, (u64 *) end - (u64 *) next);
+	k.k->u64s -= extent_entry_u64s(c, entry);
+}
+
 void bch2_bkey_extent_entry_drop(const struct bch_fs *c, struct bkey_i *k, union bch_extent_entry *entry)
 {
 	union bch_extent_entry *end = bkey_val_end(bkey_i_to_s(k));
@@ -1000,19 +1069,29 @@ void bch2_bkey_drop_ptr(const struct bch_fs *c, struct bkey_s k, struct bch_exte
 	}
 
 	bch2_bkey_drop_ptr_noerror(c, k, ptr);
+}
 
-	if (!bch2_bkey_nr_dirty_ptrs(c, k.s_c)) {
-		k.k->type = KEY_TYPE_error;
-		set_bkey_val_u64s(k.k, 0);
+void bch2_bkey_drop_ptrs_mask(const struct bch_fs *c, struct bkey_i *k, unsigned ptrs)
+{
+	while (ptrs) {
+		unsigned i = 0, drop = __fls(ptrs);
+
+		bch2_bkey_drop_ptrs_noerror(bkey_i_to_s(k), p, entry, i++ == drop);
+		ptrs  ^= 1U << drop;
 	}
 }
 
+void bch2_bkey_drop_device_noerror(const struct bch_fs *c, struct bkey_s k, unsigned dev)
+{
+	bch2_bkey_drop_ptrs_noerror(k, p, entry, p.ptr.dev == dev);
+}
+
 void bch2_bkey_drop_device(const struct bch_fs *c, struct bkey_s k, unsigned dev)
 {
 	bch2_bkey_drop_ptrs(k, p, entry, p.ptr.dev == dev);
 }
 
-void bch2_bkey_drop_ec(const struct bch_fs *c, struct bkey_i *k, unsigned dev)
+static void bch2_bkey_drop_ec(const struct bch_fs *c, struct bkey_i *k, unsigned dev)
 {
 	struct bkey_ptrs ptrs = bch2_bkey_ptrs(bkey_i_to_s(k));
 	union bch_extent_entry *entry, *ec = NULL;
@@ -1028,6 +1107,22 @@ void bch2_bkey_drop_ec(const struct bch_fs *c, struct bkey_i *k, unsigned dev)
 	}
 }
 
+void bch2_bkey_drop_ec_mask(const struct bch_fs *c, struct bkey_i *k, unsigned mask)
+{
+	while (mask) {
+		struct bkey_ptrs_c ptrs = bch2_bkey_ptrs_c(bkey_i_to_s_c(k));
+		unsigned ptr_bit = 1;
+		bkey_for_each_ptr(ptrs, ptr) {
+			if (mask & ptr_bit) {
+				bch2_bkey_drop_ec(c, k, ptr->dev);
+				mask &= ~ptr_bit;
+				break;
+			}
+			ptr_bit <<= 1;
+		}
+	}
+}
+
 const struct bch_extent_ptr *bch2_bkey_has_device_c(const struct bch_fs *c, struct bkey_s_c k, unsigned dev)
 {
 	struct bkey_ptrs_c ptrs = bch2_bkey_ptrs_c(k);
@@ -1039,13 +1134,29 @@ const struct bch_extent_ptr *bch2_bkey_has_device_c(const struct bch_fs *c, stru
 	return NULL;
 }
 
+bool bch2_bkey_has_device_decode(const struct bch_fs *c, struct bkey_s_c k, unsigned dev,
+				 struct extent_ptr_decoded *ret)
+{
+	struct bkey_ptrs_c ptrs = bch2_bkey_ptrs_c(k);
+	const union bch_extent_entry *entry;
+	struct extent_ptr_decoded p;
+
+	bkey_for_each_ptr_decode(k.k, ptrs, p, entry)
+		if (p.ptr.dev == dev) {
+			*ret = p;
+			return true;
+		}
+
+	return false;
+}
+
 bool bch2_bkey_devs_rw(struct bch_fs *c, struct bkey_s_c k)
 {
 	struct bkey_ptrs_c ptrs = bch2_bkey_ptrs_c(k);
 
 	guard(rcu)();
 	bkey_for_each_ptr(ptrs, ptr) {
-		CLASS(bch2_dev_tryget, ca)(c, ptr->dev);
+		struct bch_dev *ca = bch2_dev_rcu(c, ptr->dev);
 		if (!ca || ca->mi.state != BCH_MEMBER_STATE_rw)
 			return false;
 	}
@@ -1060,7 +1171,7 @@ bool bch2_bkey_has_target(struct bch_fs *c, struct bkey_s_c k, unsigned target)
 
 	guard(rcu)();
 	bkey_for_each_ptr(ptrs, ptr)
-		if (bch2_dev_in_target(c, ptr->dev, target) &&
+		if (bch2_dev_in_target_rcu(c, ptr->dev, target) &&
 		    (ca = bch2_dev_rcu_noerror(c, ptr->dev)) &&
 		    (!ptr->cached ||
 		     !dev_ptr_stale_rcu(ca, ptr)))
@@ -1075,12 +1186,22 @@ bool bch2_bkey_in_target(struct bch_fs *c, struct bkey_s_c k, unsigned target)
 
 	guard(rcu)();
 	bkey_for_each_ptr(ptrs, ptr)
-		if (!bch2_dev_in_target(c, ptr->dev, target))
+		if (!bch2_dev_in_target_rcu(c, ptr->dev, target))
 			return false;
 
 	return true;
 }
 
+bool bch2_bkey_has_dev_bad_or_evacuating(struct bch_fs *c, struct bkey_s_c k)
+{
+	guard(rcu)();
+	struct bkey_ptrs_c ptrs = bch2_bkey_ptrs_c(k);
+	bkey_for_each_ptr(ptrs, ptr)
+		if (bch2_dev_bad_or_evacuating_rcu(c, ptr->dev))
+			return true;
+	return false;
+}
+
 bool bch2_bkey_matches_ptr(struct bch_fs *c, struct bkey_s_c k,
 			   struct bch_extent_ptr m, u64 offset)
 {
@@ -1098,6 +1219,40 @@ bool bch2_bkey_matches_ptr(struct bch_fs *c, struct bkey_s_c k,
 	return false;
 }
 
+bool bch2_bkey_ptrs_match(struct bkey_s_c k1, struct extent_ptr_decoded p1,
+			  struct bkey_s_c k2, struct extent_ptr_decoded p2)
+{
+	return ((p1.ptr.dev		== p2.ptr.dev ||
+		 (p1.has_ec && p2.has_ec &&
+		  p1.ec.idx		== p2.ec.idx &&
+		  p1.ec.block		== p2.ec.block)) &&
+
+		p1.ptr.gen		== p2.ptr.gen &&
+
+		/*
+		 * This checks that the two pointers point
+		 * to the same region on disk - adjusting
+		 * for the difference in where the extents
+		 * start, since one may have been trimmed:
+		 */
+		(s64) p1.ptr.offset + p1.crc.offset - bkey_start_offset(k1.k) ==
+		(s64) p2.ptr.offset + p2.crc.offset - bkey_start_offset(k2.k) &&
+
+		/*
+		 * This additionally checks that the
+		 * extents overlap on disk, since the
+		 * previous check may trigger spuriously
+		 * when one extent is immediately partially
+		 * overwritten with another extent (so that
+		 * on disk they are adjacent) and
+		 * compression is in use:
+		 */
+		((p1.ptr.offset >= p2.ptr.offset &&
+		  p1.ptr.offset  < p2.ptr.offset + p2.crc.compressed_size) ||
+		 (p2.ptr.offset >= p1.ptr.offset &&
+		  p2.ptr.offset  < p1.ptr.offset + p1.crc.compressed_size)));
+}
+
 /*
  * Returns true if two extents refer to the same data:
  */
@@ -1117,31 +1272,7 @@ bool bch2_extents_match(const struct bch_fs *c, struct bkey_s_c k1, struct bkey_
 
 		bkey_for_each_ptr_decode(k1.k, ptrs1, p1, entry1)
 			bkey_for_each_ptr_decode(k2.k, ptrs2, p2, entry2)
-				if (p1.ptr.dev		== p2.ptr.dev &&
-				    p1.ptr.gen		== p2.ptr.gen &&
-
-				    /*
-				     * This checks that the two pointers point
-				     * to the same region on disk - adjusting
-				     * for the difference in where the extents
-				     * start, since one may have been trimmed:
-				     */
-				    (s64) p1.ptr.offset + p1.crc.offset - bkey_start_offset(k1.k) ==
-				    (s64) p2.ptr.offset + p2.crc.offset - bkey_start_offset(k2.k) &&
-
-				    /*
-				     * This additionally checks that the
-				     * extents overlap on disk, since the
-				     * previous check may trigger spuriously
-				     * when one extent is immediately partially
-				     * overwritten with another extent (so that
-				     * on disk they are adjacent) and
-				     * compression is in use:
-				     */
-				    ((p1.ptr.offset >= p2.ptr.offset &&
-				      p1.ptr.offset  < p2.ptr.offset + p2.crc.compressed_size) ||
-				     (p2.ptr.offset >= p1.ptr.offset &&
-				      p2.ptr.offset  < p1.ptr.offset + p1.crc.compressed_size)))
+				if (bch2_bkey_ptrs_match(k1, p1, k2, p2))
 					return true;
 
 		return false;
@@ -1169,17 +1300,41 @@ bch2_extent_has_ptr(const struct bch_fs *c, struct bkey_s_c k1,
 	return NULL;
 }
 
-static bool want_cached_ptr(struct bch_fs *c, struct bch_inode_opts *opts,
-			    struct bch_extent_ptr *ptr)
+static bool drop_cached_pointer_trace(struct bch_fs *c, struct bkey_s k,
+				      struct bch_extent_ptr *ptr, const char *msg)
 {
-	unsigned target = opts->promote_target ?: opts->foreground_target;
+	event_add_trace(c, cached_ptr_drop, k.k->size, buf, ({
+		prt_str(&buf, msg);
+		prt_str(&buf, "\ndropping pointer\n  ");
+		bch2_extent_ptr_to_text(&buf, c, ptr);
+		prt_str(&buf, "\nfrom:\n");
+		bch2_bkey_val_to_text(&buf, c, k.s_c);
+	}));
 
-	if (target && !bch2_dev_in_target(c, ptr->dev, target))
-		return false;
+	bch2_bkey_drop_ptr_noerror(c, k, ptr);
+	return true;
+}
 
-	struct bch_dev *ca = bch2_dev_rcu_noerror(c, ptr->dev);
+static bool maybe_drop_cached_ptr(struct bch_fs *c, struct bch_inode_opts *opts,
+				  struct bkey_s k,
+				  struct bch_extent_ptr *ptr, bool have_cached_ptr)
+{
+	if (ptr->cached) {
+		if (have_cached_ptr)
+			return drop_cached_pointer_trace(c, k, ptr, "extent already has another cached pointer");
 
-	return ca && bch2_dev_is_healthy(ca) && !dev_ptr_stale_rcu(ca, ptr);
+		struct bch_dev *ca = bch2_dev_rcu_noerror(c, ptr->dev);
+		if (ca && dev_ptr_stale_rcu(ca, ptr))
+			return drop_cached_pointer_trace(c, k, ptr, "pointer is stale");
+		if (!ca || ca->mi.state == BCH_MEMBER_STATE_evacuating)
+			return drop_cached_pointer_trace(c, k, ptr, "device bad or evacuating");
+
+		unsigned target = opts->promote_target ?: opts->foreground_target;
+		if (target && !bch2_dev_in_target_rcu(c, ptr->dev, target))
+			return drop_cached_pointer_trace(c, k, ptr, "incorrect target");
+	}
+
+	return false;
 }
 
 void bch2_extent_ptr_set_cached(struct bch_fs *c,
@@ -1203,20 +1358,18 @@ void bch2_extent_ptr_set_cached(struct bch_fs *c,
 			 * data. Possibly something we can fix in the future?
 			 */
 			if (&entry->ptr == ptr && p.has_ec) {
-				bch2_bkey_drop_ptr_noerror(c, k, ptr);
+				drop_cached_pointer_trace(c, k, ptr, "ptr->cached && ec currently incompatible");
 				return;
 			}
 
-			if (p.ptr.cached) {
-				if (have_cached_ptr || !want_cached_ptr(c, opts, &p.ptr)) {
-					bch2_bkey_drop_ptr_noerror(c, k, &entry->ptr);
-					ptr = NULL;
-					dropped = true;
-					break;
-				}
-
-				have_cached_ptr = true;
+			if (maybe_drop_cached_ptr(c, opts, k, &entry->ptr, have_cached_ptr)) {
+				/* @ptr is being invalidated; we'll have to find it again */
+				ptr = NULL;
+				dropped = true;
+				break;
 			}
+
+			have_cached_ptr |= p.ptr.cached;
 		}
 	} while (dropped);
 
@@ -1227,12 +1380,33 @@ void bch2_extent_ptr_set_cached(struct bch_fs *c,
 				ptr = ptr2;
 	}
 
-	if (have_cached_ptr || !want_cached_ptr(c, opts, ptr))
-		bch2_bkey_drop_ptr_noerror(c, k, ptr);
-	else
-		ptr->cached = true;
+	ptr->cached = true;
+	maybe_drop_cached_ptr(c, opts, k, ptr, have_cached_ptr);
 }
 
+static void bch2_bkey_set_ptrs_cached_mask(struct bch_fs *c,
+					   struct bch_inode_opts *opts,
+					   struct bkey_i *k,
+					   unsigned mask)
+{
+	while (mask) {
+		unsigned drop = __fls(mask);
+		mask &= ~BIT(drop);
+
+		bkey_for_each_ptr(bch2_bkey_ptrs(bkey_i_to_s(k)), ptr)
+			if (!drop--) {
+				bch2_extent_ptr_set_cached(c, opts, bkey_i_to_s(k), ptr);
+				break;
+			}
+	}
+}
+
+#if 0
+void bch2_bkey_set_ptrs_cached_mask()
+{
+}
+#endif
+
 static bool bch2_bkey_has_stale_ptrs(struct bch_fs *c, struct bkey_s_c k)
 {
 	struct bkey_ptrs_c ptrs = bch2_bkey_ptrs_c(k);
@@ -1291,45 +1465,108 @@ void bch2_bkey_drop_extra_cached_ptrs(struct bch_fs *c,
 		bool have_cached_ptr = false;
 		dropped = false;
 
-		bkey_for_each_ptr(ptrs, ptr)
-			if (ptr->cached) {
-				if (have_cached_ptr || !want_cached_ptr(c, opts, ptr)) {
-					bch2_bkey_drop_ptr_noerror(c, k, ptr);
-					dropped = true;
-					break;
-				}
-				have_cached_ptr = true;
-			}
-	} while (dropped);
-}
-
-void bch2_bkey_drop_extra_durability(struct bch_fs *c,
-				     struct bch_inode_opts *opts,
-				     struct bkey_s k)
-{
-	guard(rcu)();
-	unsigned durability = bch2_bkey_durability(c, k.s_c);
-	bool dropped;
-
-	do {
-		union bch_extent_entry *entry;
-		struct extent_ptr_decoded p;
-		dropped = false;
-
-		bkey_for_each_ptr_decode(k.k, bch2_bkey_ptrs(k), p, entry) {
-			unsigned ptr_durability = bch2_extent_ptr_durability(c, &p);
-
-			if (!p.ptr.cached &&
-			    durability - ptr_durability >= opts->data_replicas) {
-				bch2_extent_ptr_set_cached(c, opts, k, &entry->ptr);
-				durability -= ptr_durability;
+		bkey_for_each_ptr(ptrs, ptr) {
+			if (maybe_drop_cached_ptr(c, opts, k, ptr, have_cached_ptr)) {
 				dropped = true;
 				break;
 			}
+
+			have_cached_ptr |= ptr->cached;
 		}
 	} while (dropped);
 }
 
+int bch2_bkey_drop_extra_durability(struct btree_trans *trans,
+				    struct bch_inode_opts *opts,
+				    struct bkey_i *k,
+				    unsigned mask,
+				    bool kill)
+{
+	if (!mask)
+		return 0;
+
+	struct bch_fs *c = trans->c;
+	u8 ptr_durability[BCH_BKEY_PTRS_MAX];
+	unsigned durability = 0, nr_ptrs = 0;
+
+	memset(ptr_durability, 0, sizeof(ptr_durability));
+
+	union bch_extent_entry *entry;
+	struct extent_ptr_decoded p;
+
+	bkey_for_each_ptr_decode(&k->k, bch2_bkey_ptrs(bkey_i_to_s(k)), p, entry) {
+		BUG_ON(nr_ptrs >= ARRAY_SIZE(ptr_durability));
+
+		if (p.ptr.cached)
+			mask &= ~BIT(nr_ptrs);
+
+		int d = bch2_dev_idx_is_online(c, p.ptr.dev)
+			? bch2_extent_ptr_durability(trans, &p)
+			: 0;
+		if (d < 0)
+			return d;
+
+		BUG_ON(d > U8_MAX);
+
+		ptr_durability[nr_ptrs++] = d;
+		durability += d;
+	}
+
+	u8 ptrs_kill = 0;
+	for (unsigned i = 0; i < nr_ptrs; i++)
+		if ((mask & BIT(i)) &&
+		    durability - ptr_durability[i] >= opts->data_replicas) {
+			durability -= ptr_durability[i];
+			ptr_durability[i] = 0;
+			ptrs_kill |= BIT(i);
+		}
+
+	if (ptrs_kill)
+		bch2_bkey_drop_ptrs_mask(c, k, ptrs_kill);
+	else
+		bch2_bkey_set_ptrs_cached_mask(c, opts, k, ptrs_kill);
+	return 0;
+}
+
+int bch2_bkey_drop_extra_ec_durability(struct btree_trans *trans,
+				       struct bch_inode_opts *opts,
+				       struct bkey_i *k, unsigned mask)
+{
+	if (!mask)
+		return 0;
+
+	struct bkey_durability durability;
+	try(bch2_bkey_durability(trans, bkey_i_to_s_c(k), &durability));
+
+	struct bch_fs *c = trans->c;
+	union bch_extent_entry *entry;
+	struct extent_ptr_decoded p;
+	unsigned ptr_bit = 1;
+
+	bkey_for_each_ptr_decode(&k->k, bch2_bkey_ptrs(bkey_i_to_s(k)), p, entry) {
+		if (mask & ptr_bit) {
+			int d;
+			if (bch2_dev_idx_is_online(c, p.ptr.dev)) {
+				d = bch2_extent_ptr_durability(trans, &p);
+				if (d < 0)
+					return d;
+
+				d -= bch2_dev_durability(c, p.ptr.dev);
+			} else {
+				d = 0;
+			}
+
+			if (durability.online - d >= opts->data_replicas) {
+				durability.online -= d;
+				bch2_bkey_drop_ec(c, k, p.ptr.dev);
+			}
+		}
+		ptr_bit <<= 1;
+	}
+
+	return 0;
+}
+
 void bch2_extent_ptr_to_text(struct printbuf *out, struct bch_fs *c, const struct bch_extent_ptr *ptr)
 {
 	struct bch_dev *ca = c ? bch2_dev_rcu_noerror(c, ptr->dev) : NULL;
@@ -1380,10 +1617,13 @@ void bch2_bkey_ptrs_to_text(struct printbuf *out, struct bch_fs *c,
 	struct bkey_ptrs_c ptrs = bch2_bkey_ptrs_c(k);
 	const union bch_extent_entry *entry;
 
-	if (c)
-		prt_printf(out, "durability: %u ", bch2_bkey_durability_safe(c, k));
+	if (!c) {
+		prt_printf(out, "no filesystem object available");
+		return;
+	}
+
+	prt_printf(out, "durability: %u ", bch2_bkey_durability_safe(c, k));
 
-	guard(printbuf_indent)(out);
 	guard(printbuf_atomic)(out);
 	guard(rcu)();
 
@@ -1416,14 +1656,22 @@ void bch2_bkey_ptrs_to_text(struct printbuf *out, struct bch_fs *c,
 			prt_printf(out, "idx %llu block %u", (u64) ec->idx, ec->block);
 			break;
 		}
-		case BCH_EXTENT_ENTRY_rebalance:
-			bch2_extent_rebalance_to_text(out, c, &entry->rebalance);
+		case BCH_EXTENT_ENTRY_rebalance_v1:
+			bch2_extent_rebalance_v1_to_text(out, c, &entry->rebalance_v1);
+			break;
+
+		case BCH_EXTENT_ENTRY_reconcile:
+			bch2_extent_reconcile_to_text(out, c, &entry->reconcile);
 			break;
 
 		case BCH_EXTENT_ENTRY_flags:
 			prt_bitflags(out, bch2_extent_flags_strs, entry->flags.flags);
 			break;
 
+		case BCH_EXTENT_ENTRY_reconcile_bp:
+			prt_printf(out, "idx %llu", (u64) entry->reconcile_bp.idx);
+			break;
+
 		default:
 			prt_printf(out, "(unknown extent entry %.16llx)", *((u64 *) entry));
 			return;
@@ -1477,6 +1725,18 @@ fsck_err:
 	return ret;
 }
 
+static inline bool btree_ptr_entry_type_allowed(enum bch_extent_entry_type type)
+{
+	switch (type) {
+	case BCH_EXTENT_ENTRY_ptr:
+	case BCH_EXTENT_ENTRY_reconcile:
+	case BCH_EXTENT_ENTRY_reconcile_bp:
+		return true;
+	default:
+		return false;
+	};
+}
+
 int bch2_bkey_ptrs_validate(struct bch_fs *c, struct bkey_s_c k,
 			    struct bkey_validate_context from)
 {
@@ -1487,23 +1747,28 @@ int bch2_bkey_ptrs_validate(struct bch_fs *c, struct bkey_s_c k,
 	unsigned nonce = UINT_MAX;
 	unsigned nr_ptrs = 0;
 	bool have_written = false, have_unwritten = false, have_ec = false, crc_since_last_ptr = false;
+	bool have_non_inval_dev_ptrs = false;
+	bool have_non_inval_dev_ptrs_dirty = false;
 	int ret = 0;
 
 	if (bkey_is_btree_ptr(k.k))
 		size_ondisk = btree_sectors(c);
 
 	bkey_extent_entry_for_each(ptrs, entry) {
-		bkey_fsck_err_on(extent_entry_type(entry) >= c->extent_types_known,
+		unsigned type = extent_entry_type(entry);
+
+		bkey_fsck_err_on(type >= c->sb.extent_types_known,
 				 c, extent_ptrs_invalid_entry,
 				 "invalid extent entry type (got %u, max %u)",
-				 extent_entry_type(entry), c->extent_types_known);
+				 type, c->sb.extent_types_known);
 
 		bkey_fsck_err_on(bkey_is_btree_ptr(k.k) &&
-				 !extent_entry_is_ptr(entry),
+				 type < BCH_EXTENT_ENTRY_MAX &&
+				 !btree_ptr_entry_type_allowed(type),
 				 c, btree_ptr_has_non_ptr,
-				 "has non ptr field");
+				 "has non allowed field");
 
-		switch (extent_entry_type(entry)) {
+		switch (type) {
 		case BCH_EXTENT_ENTRY_ptr:
 			try(extent_ptr_validate(c, k, from, &entry->ptr, size_ondisk, false));
 
@@ -1516,6 +1781,12 @@ int bch2_bkey_ptrs_validate(struct bch_fs *c, struct bkey_s_c k,
 			else
 				have_unwritten = true;
 
+			if (entry->ptr.dev != BCH_SB_MEMBER_INVALID || have_ec) {
+				if (!entry->ptr.cached)
+					have_non_inval_dev_ptrs_dirty = true;
+				have_non_inval_dev_ptrs = true;
+			}
+
 			have_ec = false;
 			crc_since_last_ptr = false;
 			nr_ptrs++;
@@ -1566,29 +1837,16 @@ int bch2_bkey_ptrs_validate(struct bch_fs *c, struct bkey_s_c k,
 					 "redundant stripe entry");
 			have_ec = true;
 			break;
-		case BCH_EXTENT_ENTRY_rebalance: {
-			/*
-			 * this shouldn't be a fsck error, for forward
-			 * compatibility; the rebalance code should just refetch
-			 * the compression opt if it's unknown
-			 */
-#if 0
-			const struct bch_extent_rebalance *r = &entry->rebalance;
-
-			if (!bch2_compression_opt_valid(r->compression)) {
-				union bch_compression_opt opt = { .value = r->compression };
-				prt_printf(err, "invalid compression opt %u:%u",
-					   opt.type, opt.level);
-				return bch_err_throw(c, invalid_bkey);
-			}
-#endif
+		case BCH_EXTENT_ENTRY_reconcile:
+			try(bch2_extent_reconcile_validate(c, k, from, &entry->reconcile));
 			break;
-		}
 		case BCH_EXTENT_ENTRY_flags:
 			bkey_fsck_err_on(entry != ptrs.start,
 					 c, extent_flags_not_at_start,
 					 "extent flags entry not at start");
 			break;
+		case BCH_EXTENT_ENTRY_reconcile_bp:
+			break;
 		}
 	}
 
@@ -1610,6 +1868,21 @@ int bch2_bkey_ptrs_validate(struct bch_fs *c, struct bkey_s_c k,
 	bkey_fsck_err_on(have_ec,
 			 c, extent_ptrs_redundant_stripe,
 			 "redundant stripe entry");
+
+	if (!bkey_is_btree_ptr(k.k)) {
+		/*
+		 * we don't use KEY_TYPE_error for dead btree nodes - we still want the
+		 * other fields in bch_btree_ptr_v2
+		 */
+		bkey_fsck_err_on(!have_non_inval_dev_ptrs,
+				 c, extent_ptrs_all_invalid,
+				 "extent without valid pointers");
+
+		bkey_fsck_err_on(from.from == BKEY_VALIDATE_commit &&
+				 !have_non_inval_dev_ptrs_dirty,
+				 c, extent_ptrs_all_invalid,
+				 "extent without valid dirty pointers");
+	}
 fsck_err:
 	return ret;
 }
@@ -1646,7 +1919,8 @@ void bch2_ptr_swab(const struct bch_fs *c, struct bkey_s k)
 			break;
 		case BCH_EXTENT_ENTRY_stripe_ptr:
 			break;
-		case BCH_EXTENT_ENTRY_rebalance:
+		case BCH_EXTENT_ENTRY_rebalance_v1:
+		case BCH_EXTENT_ENTRY_reconcile:
 			break;
 		default:
 			/* Bad entry type: will be caught by validate() */
@@ -1720,8 +1994,10 @@ int bch2_cut_front_s(const struct bch_fs *c, struct bpos where, struct bkey_s k)
 				entry->crc128.offset += sub;
 				break;
 			case BCH_EXTENT_ENTRY_stripe_ptr:
-			case BCH_EXTENT_ENTRY_rebalance:
+			case BCH_EXTENT_ENTRY_rebalance_v1:
+			case BCH_EXTENT_ENTRY_reconcile:
 			case BCH_EXTENT_ENTRY_flags:
+			case BCH_EXTENT_ENTRY_reconcile_bp:
 				break;
 			}
 
