@@ -34,6 +34,18 @@
 #include <linux/kthread.h>
 #include <linux/sched/cputime.h>
 
+#define RECONCILE_PHASE_TYPES()		\
+	x(scan)				\
+	x(btree)			\
+	x(phys)				\
+	x(normal)			\
+
+enum reconcile_phase_type {
+#define x(n)	RECONCILE_PHASE_##n,
+	RECONCILE_PHASE_TYPES()
+#undef x
+};
+
 #define x(n) #n,
 
 const char * const bch2_reconcile_opts[] = {
@@ -50,6 +62,10 @@ static const char * const bch2_rebalance_scan_strs[] = {
 	RECONCILE_SCAN_TYPES()
 };
 
+static const char * const bch2_reconcile_phase_types[] = {
+	RECONCILE_PHASE_TYPES()
+};
+
 #undef x
 
 static bool btree_is_reconcile_phys(enum btree_id btree)
@@ -58,20 +74,6 @@ static bool btree_is_reconcile_phys(enum btree_id btree)
 		btree == BTREE_ID_reconcile_work_phys;
 }
 
-static enum reconcile_work_id btree_to_reconcile_work_id(enum btree_id btree)
-{
-	switch (btree) {
-	case BTREE_ID_reconcile_hipri:
-		return RECONCILE_WORK_hipri;
-	case BTREE_ID_reconcile_work:
-		return RECONCILE_WORK_normal;
-	case BTREE_ID_reconcile_pending:
-		return RECONCILE_WORK_pending;
-	default:
-		BUG();
-	}
-}
-
 static u64 reconcile_scan_encode(struct reconcile_scan s)
 {
 	switch (s.type) {
@@ -167,17 +169,29 @@ int bch2_set_fs_needs_reconcile(struct bch_fs *c)
 
 static int bch2_clear_reconcile_needs_scan(struct btree_trans *trans, struct bpos pos, u64 cookie)
 {
-	return commit_do(trans, NULL, NULL, BCH_TRANS_COMMIT_no_enospc, ({
-		CLASS(btree_iter, iter)(trans, BTREE_ID_reconcile_scan, pos, BTREE_ITER_intent);
-		struct bkey_s_c k = bkey_try(bch2_btree_iter_peek_slot(&iter));
+	struct bch_fs *c = trans->c;
+	u64 v;
 
-		u64 v = k.k->type == KEY_TYPE_cookie
-			? le64_to_cpu(bkey_s_c_to_cookie(k).v->cookie)
-			: 0;
-		v == cookie
-			? bch2_btree_delete_at(trans, &iter, 0)
-			: 0;
+	try(commit_do(trans, NULL, NULL, BCH_TRANS_COMMIT_no_enospc, ({
+		CLASS(btree_iter, iter)(trans, BTREE_ID_reconcile_scan, pos, BTREE_ITER_intent);
+		struct bkey_s_c k = bch2_btree_iter_peek_slot(&iter);
+		bkey_err(k) ?: ({
+			v = k.k->type == KEY_TYPE_cookie
+				? le64_to_cpu(bkey_s_c_to_cookie(k).v->cookie)
+				: 0;
+			v == cookie
+				? bch2_btree_delete_at(trans, &iter, 0)
+				: 0;
+		});
+	})));
+
+	event_inc_trace(c, reconcile_clear_scan, buf, ({
+		reconcile_scan_to_text(&buf, c, reconcile_scan_decode(c, pos.offset));
+		prt_newline(&buf);
+		prt_printf(&buf, "scan started with cookie %llu now have %llu", cookie, v);
+		prt_printf(&buf, "%sdeleting scan cookie\n", v == cookie ? "" : "not ");
 	}));
+	return 0;
 }
 
 #define RECONCILE_WORK_BUF_NR		1024
@@ -213,6 +227,8 @@ static struct bkey_s_c next_reconcile_entry(struct btree_trans *trans,
 
 		int ret = for_each_btree_key_max(trans, iter, work_pos->btree, work_pos->pos, end,
 				   flags, k, ({
+			bch2_progress_update_iter(trans, &trans->c->reconcile.progress, &iter);
+
 			/* There might be leftover scan cookies from rebalance, pre reconcile upgrade: */
 			if (k.k->type != KEY_TYPE_set)
 				continue;
@@ -720,6 +736,14 @@ static int update_reconcile_opts_scan(struct btree_trans *trans,
 					  SET_NEEDS_RECONCILE_opt_change);
 }
 
+static bool bch2_reconcile_enabled(struct bch_fs *c)
+{
+	return !c->opts.read_only &&
+		c->opts.reconcile_enabled &&
+		!(c->opts.reconcile_on_ac_only &&
+		  c->reconcile.on_battery);
+}
+
 static int do_reconcile_scan_bp(struct btree_trans *trans,
 				struct reconcile_scan s,
 				struct bkey_s_c_backpointer bp,
@@ -760,6 +784,9 @@ static int do_reconcile_scan_bps(struct moving_context *ctxt,
 				  last_flushed, NULL, bp, ({
 		ctxt->stats->pos = BBPOS(BTREE_ID_backpointers, iter.pos);
 
+		if (kthread_should_stop() || !bch2_reconcile_enabled(c))
+			break;
+
 		CLASS(disk_reservation, res)(c);
 		do_reconcile_scan_bp(trans, s, bp, last_flushed) ?:
 		bch2_trans_commit(trans, &res.r, NULL, BCH_TRANS_COMMIT_no_enospc);
@@ -821,6 +848,10 @@ static int do_reconcile_scan_btree(struct moving_context *ctxt,
 
 	return for_each_btree_key_max_continue(trans, iter, end, 0, k, ({
 		ctxt->stats->pos = BBPOS(iter.btree_id, iter.pos);
+		bch2_progress_update_iter(trans, &r->progress, &iter);
+
+		if (kthread_should_stop() || !bch2_reconcile_enabled(c))
+			return 0;
 
 		atomic64_add(!level ? k.k->size : c->opts.btree_node_size >> 9,
 			     &r->scan_stats.sectors_seen);
@@ -847,6 +878,8 @@ static int do_reconcile_scan_fs(struct moving_context *ctxt, struct reconcile_sc
 	struct bch_fs *c = ctxt->trans->c;
 	struct bch_fs_reconcile *r = &c->reconcile;
 
+	bch2_progress_init(&r->progress, NULL, c, metadata ? 0 : ~0ULL, ~0ULL);
+
 	r->scan_start	= BBPOS_MIN;
 	r->scan_end	= BBPOS_MAX;
 
@@ -929,22 +962,17 @@ static void reconcile_wait(struct bch_fs *c)
 	bch2_kthread_io_clock_wait_once(clock, r->wait_iotime_end, MAX_SCHEDULE_TIMEOUT);
 }
 
-static bool bch2_reconcile_enabled(struct bch_fs *c)
-{
-	return !c->opts.read_only &&
-		c->opts.reconcile_enabled &&
-		!(c->opts.reconcile_on_ac_only &&
-		  c->reconcile.on_battery);
-}
-
-struct reconcile_scan_phase {
-	enum btree_id	btree;
-	struct bpos	start, end;
+struct reconcile_phase {
+	enum reconcile_phase_type	type;
+	enum reconcile_work_id		priority;
+	enum btree_id			btree;
+	struct bpos			start, end;
 };
 
-static const struct reconcile_scan_phase reconcile_phases[] = {
+static const struct reconcile_phase reconcile_phases[] = {
 	/* Scan cookies: */
-	{ BTREE_ID_reconcile_scan, POS_MIN, POS(0, U64_MAX), },
+	{ RECONCILE_PHASE_scan,		RECONCILE_WORK_hipri,
+		BTREE_ID_reconcile_scan, POS_MIN, POS(0, U64_MAX), },
 
 	/* Hipri work first - evacuate/rereplicate */
 
@@ -952,28 +980,36 @@ static const struct reconcile_scan_phase reconcile_phases[] = {
 	 * Btree nodes first - they're indexed separately from the normal work
 	 * btrees because they require backpointers:
 	 */
-	{ BTREE_ID_reconcile_scan, POS(RECONCILE_WORK_hipri, 0), POS(RECONCILE_WORK_hipri, U64_MAX) },
+	{ RECONCILE_PHASE_btree,	RECONCILE_WORK_hipri,
+		BTREE_ID_reconcile_scan, POS(RECONCILE_WORK_hipri, 0), POS(RECONCILE_WORK_hipri, U64_MAX) },
 
 	/*
 	 * User data:
 	 * Phys btrees first: pending work there will also be present in the normal work btrees
 	 * Then the logical btrees, this will be data on SSDS:
 	 * */
-	{ BTREE_ID_reconcile_hipri_phys,	POS_MIN, SPOS_MAX },
-	{ BTREE_ID_reconcile_hipri,		POS_MIN, SPOS_MAX },
+	{ RECONCILE_PHASE_phys,		RECONCILE_WORK_hipri,
+		BTREE_ID_reconcile_hipri_phys,	POS_MIN, SPOS_MAX },
+	{ RECONCILE_PHASE_normal,	RECONCILE_WORK_hipri,
+		BTREE_ID_reconcile_hipri,		POS_MIN, SPOS_MAX },
 
 	/* Normal priority work: */
-	{ BTREE_ID_reconcile_scan, POS(RECONCILE_WORK_normal, 0), POS(RECONCILE_WORK_normal, U64_MAX) },
-	{ BTREE_ID_reconcile_work_phys,		POS_MIN, SPOS_MAX },
-	{ BTREE_ID_reconcile_work,		POS_MIN, SPOS_MAX },
+	{ RECONCILE_PHASE_btree,	RECONCILE_WORK_normal,
+		BTREE_ID_reconcile_scan, POS(RECONCILE_WORK_normal, 0), POS(RECONCILE_WORK_normal, U64_MAX) },
+	{ RECONCILE_PHASE_phys,		RECONCILE_WORK_normal,
+		BTREE_ID_reconcile_work_phys,		POS_MIN, SPOS_MAX },
+	{ RECONCILE_PHASE_normal,	RECONCILE_WORK_normal,
+		BTREE_ID_reconcile_work,		POS_MIN, SPOS_MAX },
 
 	/*
 	 * Lastly, work that we marked as unable to complete until system
 	 * configuration changes: this won't be process unless kicked by
 	 * something else
 	 */
-	{ BTREE_ID_reconcile_scan, POS(RECONCILE_WORK_pending, 0), POS(RECONCILE_WORK_pending, U64_MAX) },
-	{ BTREE_ID_reconcile_pending,		POS_MIN, SPOS_MAX },
+	{ RECONCILE_PHASE_btree,	RECONCILE_WORK_pending,
+		BTREE_ID_reconcile_scan, POS(RECONCILE_WORK_pending, 0), POS(RECONCILE_WORK_pending, U64_MAX) },
+	{ RECONCILE_PHASE_normal,	RECONCILE_WORK_pending,
+		BTREE_ID_reconcile_pending,		POS_MIN, SPOS_MAX },
 };
 
 typedef struct {
@@ -1059,15 +1095,29 @@ static int do_reconcile_phys(struct bch_fs *c, unsigned reconcile_phase)
 	return 0;
 }
 
-static struct bbpos reconcile_phase_start(unsigned i)
+static void reconcile_phase_start(struct bch_fs *c)
 {
-	struct reconcile_scan_phase p = reconcile_phases[i];
-	return BBPOS(p.btree, p.start);
+	struct bch_fs_reconcile *r = &c->reconcile;
+	struct reconcile_phase p = reconcile_phases[r->phase];
+	r->work_pos = BBPOS(p.btree, p.start);
+
+	switch (p.type) {
+	case RECONCILE_PHASE_normal:
+		bch2_progress_init(&r->progress, NULL, c,
+				   BIT_ULL(reconcile_work_btree[p.priority]), 0);
+		break;
+	case RECONCILE_PHASE_phys:
+		bch2_progress_init(&r->progress, NULL, c,
+				   BIT_ULL(reconcile_work_phys_btree[p.priority]), 0);
+		break;
+	default:
+		break;
+	}
 }
 
 static bool reconcile_phase_is_pending(unsigned i)
 {
-	struct reconcile_scan_phase p = reconcile_phases[i];
+	struct reconcile_phase p = reconcile_phases[i];
 	return (p.btree == BTREE_ID_reconcile_scan &&
 		p.start.inode == RECONCILE_WORK_pending) ||
 		p.btree == BTREE_ID_reconcile_pending;
@@ -1090,9 +1140,8 @@ static int do_reconcile(struct moving_context *ctxt)
 
 	CLASS(per_snapshot_io_opts, snapshot_io_opts)(c);
 
-	unsigned i = 0;
-
-	r->work_pos = reconcile_phase_start(i);
+	r->phase = 0;
+	reconcile_phase_start(c);
 
 	struct bkey_i_cookie pending_cookie;
 	bkey_init(&pending_cookie.k);
@@ -1115,28 +1164,27 @@ static int do_reconcile(struct moving_context *ctxt)
 
 		if (kick != r->kick) {
 			kick		= r->kick;
-			i		= 0;
-			r->work_pos	= BBPOS(reconcile_phases[i].btree,
-						reconcile_phases[i].start);
 			work.nr		= 0;
+			r->phase	= 0;
+			reconcile_phase_start(c);
 		}
 
 		bch2_trans_begin(trans);
 
 		struct bkey_s_c k = next_reconcile_entry(trans, &work,
 							 &r->work_pos,
-							 reconcile_phases[i].end);
+							 reconcile_phases[r->phase].end);
 		ret = bkey_err(k);
 		if (ret)
 			break;
 
 		if (!k.k) {
-			if (++i == ARRAY_SIZE(reconcile_phases))
+			if (++r->phase == ARRAY_SIZE(reconcile_phases))
 				break;
 
-			r->work_pos = reconcile_phase_start(i);
+			reconcile_phase_start(c);
 
-			if (reconcile_phase_is_pending(i) &&
+			if (reconcile_phase_is_pending(r->phase) &&
 			    bkey_deleted(&pending_cookie.k))
 				break;
 
@@ -1158,13 +1206,28 @@ static int do_reconcile(struct moving_context *ctxt)
 						k.k->p,
 						le64_to_cpu(bkey_s_c_to_cookie(k).v->cookie),
 						&sectors_scanned, &last_flushed);
+
+			if (bch2_err_matches(ret, BCH_ERR_transaction_restart)) {
+#ifdef CONFIG_BCACHEFS_DEBUG
+				CLASS(printbuf, buf)();
+				bch2_prt_backtrace(&buf, &trans->last_restarted_trace);
+				panic("in transaction restart: %s, last restarted by\n%s",
+				      bch2_err_str(trans->restarted),
+				      buf.buf);
+#else
+				panic("in transaction restart: %s, last restarted by %pS\n",
+				      bch2_err_str(trans->restarted),
+				      (void *) trans->last_restarted_ip);
+#endif
+			}
 		} else if (k.k->type == KEY_TYPE_backpointer) {
 			ret = do_reconcile_btree(ctxt, &snapshot_io_opts,
 						 r->work_pos, bkey_s_c_to_backpointer(k));
 		} else if (btree_is_reconcile_phys(r->work_pos.btree)) {
 			bch2_trans_unlock_long(trans);
-			ret = do_reconcile_phys(c, i);
-			r->work_pos = reconcile_phase_start(++i);
+			ret = do_reconcile_phys(c, r->phase);
+			BUG_ON(bch2_err_matches(ret, BCH_ERR_transaction_restart));
+			reconcile_phase_start(c);
 		} else {
 			ret = lockrestart_do(trans,
 				do_reconcile_extent(ctxt, &snapshot_io_opts, r->work_pos));
@@ -1250,25 +1313,6 @@ void bch2_reconcile_status_to_text(struct printbuf *out, struct bch_fs *c)
 
 	struct bch_fs_reconcile *r = &c->reconcile;
 
-	prt_printf(out, "pending work:\tdata\rmetadata\r\n");
-	for (unsigned i = 0; i < BCH_RECONCILE_ACCOUNTING_NR; i++) {
-		struct disk_accounting_pos acc;
-		disk_accounting_key_init(acc, reconcile_work, i);
-		u64 v[2];
-		bch2_accounting_mem_read(c, disk_accounting_pos_to_bpos(&acc), v, ARRAY_SIZE(v));
-
-		bch2_prt_reconcile_accounting_type(out, i);
-		prt_printf(out, ":\t");
-		prt_human_readable_u64(out, v[0] << 9);
-		prt_tab_rjust(out);
-		prt_human_readable_u64(out, v[1] << 9);
-		prt_tab_rjust(out);
-		prt_newline(out);
-	}
-
-	prt_newline(out);
-	guard(printbuf_indent_nextline)(out);
-
 	if (!r->running) {
 		prt_printf(out, "waiting:\n");
 		u64 now = atomic64_read(&c->io_clock[WRITE].now);
@@ -1285,30 +1329,32 @@ void bch2_reconcile_status_to_text(struct printbuf *out, struct bch_fs *c)
 		bch2_pr_time_units(out, ktime_get_real_ns() - r->wait_wallclock_start);
 		prt_newline(out);
 	} else {
-		struct bbpos work_pos = r->work_pos;
+		struct reconcile_phase phase = reconcile_phases[r->phase];
+		struct bpos work_pos = r->work_pos.pos;
 		barrier();
 
-		if (work_pos.btree	== BTREE_ID_reconcile_scan &&
-		    work_pos.pos.inode	== 0) {
-			prt_printf(out, "scanning:\n");
-			reconcile_scan_to_text(out, c,
-				reconcile_scan_decode(c, work_pos.pos.offset));
-		} else if (work_pos.btree == BTREE_ID_reconcile_scan) {
-			prt_printf(out, "processing metadata: %s %llu\n",
-				   bch2_reconcile_work_ids[work_pos.pos.inode - 1],
-				   work_pos.pos.offset);
+		if (phase.type == RECONCILE_PHASE_scan) {
+			prt_printf(out, "scanning: ");
+			struct reconcile_scan s = reconcile_scan_decode(c, work_pos.offset);
+			reconcile_scan_to_text(out, c, s);
 
-		} else {
-			prt_printf(out, "processing data: ");
-
-			if (btree_is_reconcile_phys(work_pos.btree)) {
-				bch2_bbpos_to_text(out, work_pos);
-			} else {
-				prt_printf(out, " %s ",
-					   bch2_reconcile_work_ids[btree_to_reconcile_work_id(work_pos.btree)]);
-
-				bch2_bbpos_to_text(out, rb_work_to_data_pos(work_pos.pos));
+			if (s.type == RECONCILE_SCAN_fs ||
+			    s.type == RECONCILE_SCAN_metadata) {
+				prt_char(out, ' ');
+				bch2_progress_to_text(out, &r->progress);
 			}
+			prt_newline(out);
+		} else {
+			prt_printf(out, "processing %s %s: ",
+				   bch2_reconcile_work_ids[phase.priority],
+				   bch2_reconcile_phase_types[phase.type]);
+
+			if (phase.type == RECONCILE_PHASE_normal) {
+				bch2_progress_to_text(out, &r->progress);
+			} else {
+				bch2_bpos_to_text(out, work_pos);
+			}
+
 			prt_newline(out);
 		}
 	}
@@ -1320,9 +1366,15 @@ void bch2_reconcile_status_to_text(struct printbuf *out, struct bch_fs *c)
 			get_task_struct(t);
 	}
 
+	prt_newline(out);
+
 	if (t) {
+		prt_str(out, "Reconcile thread backtrace:\n");
+		guard(printbuf_indent)(out);
 		bch2_prt_task_backtrace(out, t, 0, GFP_KERNEL);
 		put_task_struct(t);
+	} else {
+		prt_str(out, "Reconcile thread not running\n");
 	}
 }
 
