@@ -197,7 +197,7 @@ void bch2_btree_node_ondisk_to_text(struct printbuf *out, struct bch_fs *c,
 	unsigned offset = 0;
 	int ret;
 
-	if (bch2_bkey_pick_read_device(c, bkey_i_to_s_c(&b->key), NULL, &pick, -1) <= 0) {
+	if (bch2_bkey_pick_read_device(c, bkey_i_to_s_c(&b->key), NULL, &pick, 0, 0) <= 0) {
 		prt_printf(out, "error getting device to read from: invalid device\n");
 		return;
 	}
@@ -953,7 +953,8 @@ static ssize_t bch2_subvolumes_read(struct file *file, char __user *buf,
 {
 	struct dump_iter *i = file->private_data;
 
-	if (!test_bit(BCH_FS_may_go_rw, &i->c->flags))
+	if (!test_bit(BCH_FS_may_go_rw, &i->c->flags) ||
+	    !i->c->snapshots.table)
 		return 0;
 
 	return bch2_simple_print(file, buf, size, ppos, bch2_subvolumes_list_to_text);
@@ -971,7 +972,8 @@ static ssize_t bch2_snapshot_trees_read(struct file *file, char __user *buf,
 {
 	struct dump_iter *i = file->private_data;
 
-	if (!test_bit(BCH_FS_may_go_rw, &i->c->flags))
+	if (!test_bit(BCH_FS_may_go_rw, &i->c->flags) ||
+	    !i->c->snapshots.table)
 		return 0;
 
 	return bch2_simple_print(file, buf, size, ppos, bch2_snapshot_trees_to_text);
