@@ -344,7 +344,7 @@ int __bch2_inode_peek(struct btree_trans *trans,
 		      struct btree_iter *iter,
 		      struct bch_inode_unpacked *inode,
 		      subvol_inum inum, unsigned flags,
-		      bool warn)
+		      const char *warn)
 {
 	u32 snapshot;
 	try(__bch2_subvolume_get_snapshot(trans, inum.subvol, &snapshot, warn));
@@ -367,7 +367,8 @@ int __bch2_inode_peek(struct btree_trans *trans,
 	return 0;
 err:
 	if (warn)
-		bch_err_msg(trans->c, ret, "looking up inum %llu:%llu:", inum.subvol, inum.inum);
+		bch_err_msg(trans->c, ret, "%s(): looking up inum %llu:%llu:",
+			    warn, inum.subvol, inum.inum);
 	return ret;
 }
 
@@ -387,7 +388,7 @@ int bch2_inode_find_by_inum_snapshot(struct btree_trans *trans,
 int __bch2_inode_find_by_inum_trans(struct btree_trans *trans,
 				    subvol_inum inum,
 				    struct bch_inode_unpacked *inode,
-				    bool warn)
+				    const char *warn)
 {
 	CLASS(btree_iter_uninit, iter)(trans);
 	return __bch2_inode_peek(trans, &iter, inode, inum, 0, warn);
@@ -412,7 +413,7 @@ int bch2_inode_find_oldest_snapshot(struct btree_trans *trans, u64 inum, u32 sna
 		if (k.k->p.offset != inum)
 			break;
 		if (!bkey_is_inode(k.k) ||
-		    !bch2_snapshot_is_ancestor(trans->c, snapshot, k.k->p.snapshot))
+		    !bch2_snapshot_is_ancestor(trans, snapshot, k.k->p.snapshot))
 			continue;
 		try(bch2_inode_unpack(k, root));
 		ret = 0;
@@ -658,7 +659,6 @@ static struct bkey_s_c
 bch2_bkey_get_iter_snapshot_parent(struct btree_trans *trans, struct btree_iter *iter,
 				   enum btree_id btree, struct bpos pos)
 {
-	struct bch_fs *c = trans->c;
 	struct bkey_s_c k;
 	int ret = 0;
 
@@ -667,7 +667,7 @@ bch2_bkey_get_iter_snapshot_parent(struct btree_trans *trans, struct btree_iter
 
 	for_each_btree_key_max_continue_norestart(*iter, SPOS(pos.inode, pos.offset, U32_MAX),
 						  BTREE_ITER_all_snapshots, k, ret)
-		if (bch2_snapshot_is_ancestor(c, pos.snapshot, k.k->p.snapshot))
+		if (bch2_snapshot_is_ancestor(trans, pos.snapshot, k.k->p.snapshot))
 			return k;
 
 	return ret ? bkey_s_c_err(ret) : bkey_s_c_null;
@@ -691,14 +691,13 @@ bch2_inode_get_iter_snapshot_parent(struct btree_trans *trans, struct btree_iter
 
 int __bch2_inode_has_child_snapshots(struct btree_trans *trans, struct bpos pos)
 {
-	struct bch_fs *c = trans->c;
 	struct bkey_s_c k;
 	int ret = 0;
 
 	for_each_btree_key_max_norestart(trans, iter,
 			BTREE_ID_inodes, POS(0, pos.offset), bpos_predecessor(pos),
 			BTREE_ITER_all_snapshots, k, ret)
-		if (bch2_snapshot_is_ancestor(c, k.k->p.snapshot, pos.snapshot) &&
+		if (bch2_snapshot_is_ancestor(trans, k.k->p.snapshot, pos.snapshot) &&
 		    bkey_is_inode(k.k)) {
 			ret = 1;
 			break;
@@ -968,7 +967,7 @@ int bch2_inode_create(struct btree_trans *trans,
 			if (pos < iter->pos.offset)
 				break;
 
-			if (bch2_snapshot_is_ancestor(trans->c, snapshot, k.k->p.snapshot) &&
+			if (bch2_snapshot_is_ancestor(trans, snapshot, k.k->p.snapshot) &&
 			    k.k->type == KEY_TYPE_inode_generation) {
 				pos = k.k->p.offset;
 				gen = le32_to_cpu(bkey_s_c_to_inode_generation(k).v->bi_generation);
@@ -1173,7 +1172,7 @@ void bch2_inode_opts_get_inode(struct bch_fs *c,
 	BCH_INODE_OPTS()
 #undef x
 
-	ret->change_cookie = atomic_read(&c->opt_change_cookie);
+	ret->change_cookie = c->opt_change_cookie;
 
 	bch2_io_opts_fixups(ret);
 }
