@@ -371,19 +371,11 @@ int bch2_repair_inode_hash_info(struct btree_trans *,
 				struct bch_inode_unpacked *);
 
 struct snapshots_seen;
-int bch2_str_hash_repair_key(struct btree_trans *,
-			     struct snapshots_seen *,
-			     const struct bch_hash_desc *,
-			     struct bch_hash_info *,
-			     struct btree_iter *, struct bkey_s_c,
-			     struct btree_iter *, struct bkey_s_c,
-			     bool *);
-
 int __bch2_str_hash_check_key(struct btree_trans *,
 			      struct snapshots_seen *,
 			      const struct bch_hash_desc *,
 			      struct bch_hash_info *,
-			      struct btree_iter *, struct bkey_s_c,
+			      struct bkey_s_c,
 			      bool *, bool *);
 
 static inline bool str_hash_key_needs_check(const struct bch_hash_desc *desc,
@@ -411,12 +403,12 @@ static inline int bch2_str_hash_check_key(struct btree_trans *trans,
 			    struct snapshots_seen *s,
 			    const struct bch_hash_desc *desc,
 			    struct bch_hash_info *hash_info,
-			    struct btree_iter *k_iter, struct bkey_s_c hash_k,
+			    struct bkey_s_c hash_k,
 			    bool *updated_before_k_pos,
 			    bool *repaired_inode)
 {
 	return str_hash_key_needs_check(desc, hash_info, hash_k)
-		? __bch2_str_hash_check_key(trans, s, desc, hash_info, k_iter, hash_k,
+		? __bch2_str_hash_check_key(trans, s, desc, hash_info, hash_k,
 					    updated_before_k_pos, repaired_inode)
 		: 0;
 }
