@@ -532,7 +532,6 @@ static int __journal_res_get(struct journal *j, struct journal_res *res,
 {
 	struct bch_fs *c = container_of(j, struct bch_fs, journal);
 	struct journal_buf *buf;
-	bool can_discard;
 	int ret;
 retry:
 	if (journal_res_get_fast(j, res, flags))
@@ -547,7 +546,6 @@ retry:
 
 	if ((flags & BCH_WATERMARK_MASK) < j->watermark) {
 		ret = bch_err_throw(c, journal_full);
-		can_discard = j->can_discard;
 		goto out;
 	}
 
@@ -584,7 +582,6 @@ retry:
 	__journal_entry_close(j, JOURNAL_ENTRY_CLOSED_VAL, false);
 	ret = journal_entry_open(j) ?: -BCH_ERR_journal_retry_open;
 unlock:
-	can_discard = j->can_discard;
 	spin_unlock(&j->lock);
 out:
 	if (likely(!ret))
@@ -595,11 +592,17 @@ out:
 	if (journal_error_check_stuck(j, ret, flags))
 		ret = bch_err_throw(c, journal_stuck);
 
-	if ((ret == -BCH_ERR_journal_max_in_flight &&
-	     track_event_change(&c->times[BCH_TIME_blocked_journal_max_in_flight], true)) ||
-	    (ret == -BCH_ERR_journal_max_open &&
-	     track_event_change(&c->times[BCH_TIME_blocked_journal_max_open], true)))
-		event_inc_trace(c, journal_entry_full, buf, ({
+	if (ret == -BCH_ERR_journal_blocked)
+		track_event_change(&c->times[BCH_TIME_blocked_journal_write_buffer_flush], true);
+
+	if (ret == -BCH_ERR_journal_max_in_flight)
+		track_event_change(&c->times[BCH_TIME_blocked_journal_max_in_flight], true);
+
+	if (ret == -BCH_ERR_journal_max_open)
+		track_event_change(&c->times[BCH_TIME_blocked_journal_max_open], true);
+
+	if (bch2_err_matches(ret, BCH_ERR_operation_blocked))
+		event_inc_trace(c, journal_res_get_blocked, buf, ({
 			prt_printf(&buf, "%s\n", bch2_err_str(ret));
 			bch2_printbuf_make_room(&buf, 4096);
 
@@ -616,11 +619,6 @@ out:
 	if ((ret == -BCH_ERR_journal_full ||
 	     ret == -BCH_ERR_journal_pin_full) &&
 	    !(flags & JOURNAL_RES_GET_NONBLOCK)) {
-		if (can_discard) {
-			bch2_journal_do_discards(j);
-			goto retry;
-		}
-
 		if (journal_low_on_space(j) &&
 		    mutex_trylock(&j->reclaim_lock)) {
 			bch2_journal_reclaim(j);
@@ -945,6 +943,9 @@ void bch2_journal_unblock(struct journal *j)
 				new.v = old.v;
 				new.cur_entry_offset = j->cur_entry_offset_if_blocked;
 			} while (!atomic64_try_cmpxchg(&j->reservations.counter, &old.v, new.v));
+
+			struct bch_fs *c = container_of(j, struct bch_fs, journal);
+			track_event_change(&c->times[BCH_TIME_blocked_journal_write_buffer_flush], false);
 		}
 
 	journal_wake(j);
@@ -1137,7 +1138,7 @@ void __bch2_journal_debug_to_text(struct printbuf *out, struct journal *j)
 		prt_printf(out, "cur_idx\t%u (seq %llu)\n",	ja->cur_idx,		ja->bucket_seq[ja->cur_idx]);
 	}
 
-	prt_printf(out, "replicas want %u need %u\n", c->opts.metadata_replicas, c->opts.metadata_replicas_required);
+	prt_printf(out, "replicas %u\n", c->opts.metadata_replicas);
 }
 
 void bch2_journal_debug_to_text(struct printbuf *out, struct journal *j)
