@@ -66,18 +66,16 @@ DEFINE_CLASS(_type, _type, darray_exit(&(_T)), (_type) {}, void)
 #define DEFINE_DARRAY_CLASS_FREE_ITEM(_type, _free)		\
 DEFINE_CLASS(_type, _type, darray_exit_free_item(&(_T), _free), (_type) {}, void)
 
-#define DEFINE_DARRAY(_type)						\
-typedef DARRAY(_type)	darray_##_type;					\
-DEFINE_DARRAY_CLASS(darray_##_type)
+#define DEFINE_DARRAY_NAMED(_name, _type)				\
+typedef DARRAY(_type)	_name;						\
+DEFINE_DARRAY_CLASS(_name)
+
+#define DEFINE_DARRAY(_type)	DEFINE_DARRAY_NAMED(darray_##_type, _type)
 
 #define DEFINE_DARRAY_PREALLOCATED(_type, _nr)				\
 typedef DARRAY_PREALLOCATED(_type, _nr)	darray_##_type;			\
 DEFINE_DARRAY_CLASS(darray_##_type)
 
-#define DEFINE_DARRAY_NAMED(_name, _type)				\
-typedef DARRAY(_type)	_name;						\
-DEFINE_DARRAY_CLASS(_name)
-
 #define DEFINE_DARRAY_NAMED_FREE_ITEM(_name, _type, _free)		\
 typedef DARRAY(_type)	_name;						\
 DEFINE_DARRAY_CLASS_FREE_ITEM(_name, _free)
@@ -171,6 +169,16 @@ int __bch2_darray_resize_noprof(darray_char *, size_t, size_t, gfp_t, bool);
 #define darray_find(_d, _item)	darray_find_p(_d, _i, *_i == _item)
 
 #define darray_sort(_d, _cmp)						\
-	sort((_d).data, (_d).nr, sizeof((_d).data[0]), _cmp, NULL)
+	sort_nonatomic((_d).data, (_d).nr, sizeof((_d).data[0]), _cmp, NULL)
+
+#define darray_eytzinger1_sort(_d, _cmp)				\
+	eytzinger1_sort((_d).data, (_d).nr - 1, sizeof((_d).data[0]), _cmp, NULL)
+
+#define darray_eytzinger1_find(_d, _cmp, _search)				\
+({										\
+	int idx = eytzinger1_find((_d).data, (_d).nr - 1, sizeof((_d).data[0]),	\
+				  _cmp, _search);				\
+	idx ? (_d).data + idx : NULL;						\
+})
 
 #endif /* _BCACHEFS_DARRAY_H */
